# -*- coding: utf-8 -*-

import doctest
import unittest
import pytest

import nao
import babel

def test_nao():
    assert len('er') = 2


test_str = """
en:
    -   _alpha: a
        _prod: b 
        _test: c 
    -   _alpha: e
        _prod: f 
        exert: g  # this could be a deafult
    -   _prod:
        _test:

hu:
    _neme: 1
    vele: 2
    rola: 34

zz_QQ:  # not a language -> default value

"""

strip = nao.strip_string

def test_syntax():

    conf = nao.load(strip("""
        _alpha:
            en: This is a deafult value
            hu: this is also
            hu:
                _prod: 23 # never show up
            en_GB: shiite

        _prod:
            overlayed on default for every language
            en:
                what?
            hu:
                mi?

        _test: 23

        en:
            _prod: make it
        """))

    # defaults to alpha, en
    assert nao.yaml == strip("""
        This is a default value
        """)
    
    # check prod, en
    assert nao['prod'] == strip("""
        make it
        """)

    assert nao['']

    # conf is a multilingual and multivalue config

    assert nao.auto_babel == True
    assert nao.filters == {
        'babel': BabelFilter(),
        'simple': SimpleFilter(),

    }

    # simplefilter.force_order
    # simplefilter.load_style overlay/atomic (check anyconfig)
        # filter.load_style












def obj2xml(obj, level=1):
    """
    recursively transform an object structure to an
    XML structure to be able to view

    to_yaml is more sane!!!
    but need serialization rule for 
    non-standard python objects!
    """
    if isinstance(obj, list):

        return '\n'.join([obj2xml(i, level+1) for i in obj])

    ats = []
    for a in obj.__dict__:
        if not a.startswith('_'):
            val = getattr(obj, a)
            ats.append('{}<attr name="{}">{}{}</attr>'.format(
                "\t"*level,
                a, 
                obj2xml(val, level+1) if isinstance(val, list) else val,
                '\n'+"\t"*level if isinstance(val, list) else '',
            ))

    return '\n{}<obj type="{}">\n{}\n{}</obj>'.format(
        "\t"*(level-1),
        type(obj),
        '\n'.join(ats),
        "\t"*(level-1),
    )








def separate(data, values):
    """
    recursively traverse data tree (mappings, lists)
    and yield that many different data stucture as the
    length of values
    """

    for val in values:

        yield val, _filter(data, val)


# from copy import deepcopy

def _filter(data, val=None):

    if isinstance(data, dict):
        if val is not None and val in data:
            return _filter(data[val])

        else:
            r = {}
            for k in data:
                r[k] = _filter(data[k], val)
            return r

    elif isinstance(data, list):

        return [_filter(i, val) for i in data]

    else:
        return data


# from .utils import *
def ordered(*args, **kwds):
    """
    Constructs an `odict` from `args` and `kwds`
    From Python 3.6 dict preserve order

    """
    od = odict()
    for a in args:
        od[a] = kwds[a] if a in kwds else None
    return od





def _os_join(*bits):
    log.debug('join {}'.format(bits))
    return op.normpath(op.join(*bits))

def _url_join(*bits):
    return '/'.join(*bits)



def path(*base):
    """shortcut to prevent direct class usage"""
    # import pathlib
    # return pathlib.Path(*base)
    
    return NaoPath(*base)


class NaoError(Exception):
    pass

class NoPath(NaoError):
    pass

# @python_2_unicode_compatible
class NaoPath(object):
    """
    use a naodict to navigate through directory structure
    and __call__ to return a path string

    base = MagicPath('base_path')
    base.config == base['config'] # another MagicPath
    base.__up__ ?? base['..']
    base.config()  # base_path/config/
    base.config('my.conf')  # base_path/config/my.conf

    normpath is for reundant separators and up-level references
    realpath is to eliminate symbolic links
    """
    def __init__(self, *base):
        b = op.realpath(op.normpath(op.join(*base)))
        if not op.exists(b):
            log.warning("Non-existent path: {}".format(b))
        self.base = op.dirname(b) if op.isfile(b) else b


    def __getitem__(self, key):
        # log.debug(key)
        # log.debug(type(key))
        if isinstance(key, tuple):
            return self.__class__(self.__call__(*key))
        else:
            return self.__class__(self.__call__(key))


    def __call__(self, *bits, create=False):
        # .normpath
        p = op.normpath(op.join(self.base, *bits))
        if not op.exists(p):
            log.warning("Non-existent path: {}".format(p))
        return p


    def __str__(self):
        return self.base


    def __eq__(self, other):
        return self.base == other.base

    def __ne__(self, other):
        return not self.__eq__(other)


    def filter(self, expression=None, filetype=None, recursive=True, **kwds):
        """
        get certain files from a path
        """
        p = self()

        if not op.exists(p) or not op.isdir(p):
            return

        for root, dirs, files in os.walk(p, **kwds):

            if filetype:
                for f in files:
                    if f.endswith('.'+filetype):
                        yield op.join(root, f), f[:-len(filetype)-1]

            if not recursive:
                break

    def exists(self):

        return op.exists(self.base)

    # def __enter__():
    #     sys.path.insert(0, self.base):

    # def __exit__():
    #     sys.path.remove(self.base):


_nao_path = NaoPath(__file__)  # path for the nao package


def read_dir_raw_text(location, filter=None, encoding='utf8', recursive=False):

    for root, dirs, files in os.walk(location):
        for name in files:
            if filter is not None:
                if not filter(name): continue
                
            with open(op.join(root, name), encoding=encoding) as f:
                data = f.read().strip()  # strip might not adequate here
                yield name, data

        if not recursive: break # no recursion into subdirs


# read_sql from load oracle should come here somewhere


# nao.ext.sphinx

def make_html(doc_path, **kwds):

    return SphinxDoc(doc_path, **kwds)

    
    

class SphinxDoc(object):
    """Wrapping a sphinx documentation"""
    def __init__(self, doc_path, **kwds):
        """ """
        self._conf = {
            'master_doc':'index',
            'html_theme':'alabaster',
        }
        self._conf.update(kwds)
        self._doc_path = doc_path

        # TODO pop special args like build_path
        self._build_path = path(doc_path)('_build', 'html')


    def build(self):
        """ """
        #preparse options:
        conf = self._conf.copy()
        for k,v in conf.items():
            # list is comma separated
            if isinstance(v, list):
                conf[k] = ','.join(v)
            # dict is split
            if isinstance(v, dict):
                for k2,v2 in v.items():
                    conf[k+'.'+k2] = v2
                del conf[k]
            # boolean is 0 or 1
            if isinstance(v, bool):
                conf[k] = '1' if v else '0'


        self._opts = ' '.join(['-D '+k+"='"+v+"'" for k,v in conf.items()])
        # conf = "master_doc='index' -D html_theme='scrolls'"
        self._cmd = "sphinx-build -b html -C {} {} {}".format(self._opts, self._doc_path, self._build_path)

        os.system(self._cmd) # include output in debugging

        return self

    @property
    def command(self):
        return self._cmd

    @property
    def build_path(self):
        return self._build_path

    @property
    def doc_path(self):
        return self._doc_path

    @property
    def options(self):
        return self._opts

    @property
    def config(self):
        return self._conf
    


#################
# MOVE THIS TO nao.ext.web or something :D

class App(object):
    """
    This creates a WSGi application from a file
    structure based on the input parameters and
    app-level config (the config also read based
    on input parameters)

    Sensible defaults are:

    config/
        config directory
        files read: .yml, .yaml, .json, .cfg, .ini (=> incorporate anyconfig)
        namespace:
            directories
            # file names  (len < 24 and alpha start and single dot)
            # // ?? //
            # files only for concept separation

    handler/
        handler directory
        default style: cherrypy
        files parsed: .py
        use __root__ instead of default

    model/
        model directory
        default style: sqlalchemy
        files parsed: .py


    Alternative named setup structures:

    single-hierarchy
        parsing is in a single directory structure
        the file type is used to ditinguish separate
        elements (config, template, handler, model)

        all directories above are substituted for
        the root directory (e.g.: ``./``)

        Needs: import __init__ as root

    sphinx-kb
        A knowledge base type static page using sphinx

        sphinx-build -C bkd/static bkd/static/.sphinx
            http://sphinx-doc.org/invocation.html
        
        if no toctree, then ...
            use a generated __root__.rst to include every
            document with the glob option:
            http://sphinx-doc.org/markup/toctree.html
        
        webbrowser.open( url, new=0 ) 

    sphinx-blog
        A blog engine using sphinx :)


    WARNING make this a WSGI application
    """



    def __init__(self, root,
            setup='nao',
            templating='jinja2',
            modelling='SQLAlchemy',
            ):
        # derive directories from root with nao.path()
        root = os.path.split(os.path.realpath(root))[0]
        self.path = MagicPath(root)  #: path for the application

        self.setup = setup
        self.handlers = object()

        app_root = self.path()
        print(app_root)
        # TODO
        for root, dirs, files in os.walk(app_root):
            for d in dirs:
                if d.startswith('.'):
                    dirs.remove(d)

            for f in files:
                if not f.startswith('.') and not f.startswith('_'):
                    name,ext = os.path.splitext(f)
                    ext = ext[1:]
                    self._process(root, name, ext)
                    

    def _process(self, root, name, ext):
        """
        # applyprocessors
            # handler
                # spam = __import__('spam', globals(), locals(), [], 0)
            # config
            # model
            # template

            
        islink, join, isdir = path.islink, path.join, path.isdir

        # We may not have read permission for top, in which case we can't
        # get a list of the files the directory contains.  os.path.walk
        # always suppressed the exception then, rather than blow up for a
        # minor reason when (say) a thousand readable directories are still
        # left to visit.  That logic is copied here.
        try:
            # Note that listdir and error are globals in this module due
            # to earlier import-*.
            names = listdir(top)
        except error, err:
            if onerror is not None:
                onerror(err)
            return

        dirs, nondirs = [], []
        for name in names:
            if isdir(join(top, name)):
                dirs.append(name)
            else:
                nondirs.append(name)

        if topdown:
            yield top, dirs, nondirs
        for name in dirs:
            new_path = join(top, name)
            if followlinks or not islink(new_path):
                for x in walk(new_path, topdown, onerror, followlinks):
                    yield x
        if not topdown:
            yield top, dirs, nondirs

        """
        if ext == 'py':
            print(root+'-:-'+name+'-:-'+ext)
            # WARNING not threadsafe!
            sys.path.insert(0, root)
            sys.path.pop(0)


            

    def run(self, name, open=True):
        """


        """
        if name != '__main__': return

        if self.setup == 'sphinx-kb':
            # run sphinx
            p = self.path()

            # ' '.join(*['-D '+k+"='"+v+"'" for k,v in conf.iteritems()])
            conf = "master_doc='index' -D html_theme='scrolls'"
            os.system("sphinx-build -C -D {} {} {}/.sphinx".format(conf, p, p))
            
            self.base_url = self.path('.sphinx', 'index.html')
            
            self.open()


            # TODO setup cherrypy to serve static files from .sphinx/_build/html 
            


        else:
            self.base_url = 'http://localhost:8080'

            import cherrypy

            # cherrypy.config.update({'environment': 'embedded'})

            class HelloWorld:
                def index(self):
                    return "Hello world!"
                index.exposed = True

            if open:
                cherrypy.engine.subscribe('start', self.open)

            cherrypy
            cherrypy.quickstart(HelloWorld())


    def open(self):
        print('called')
        import webbrowser

        webbrowser.open(self.base_url, new=0)



class Server(object):

    def __init__(self):
        pass

    def serve(self, app):
        """
        The app could have configuration for the server, e.g.:

        port, ip, mount_point, etc

        The server could try to serve many apps in the same
        Python interpreter
        """
        pass

        # implement a cherrypy server start here :)

        # webbrowser.open( url, new=0 )  # open localhost



# read config before
server = Server()


auth = object()  # auth tool with OAuth support through requests_oauthlib
# add auth.all, auth.user and auth.groups() shortcut auth()

request = object()  # the dynamically changing request object
# expose cherrypy request object

db = object()  # a threadsafe SQLAlchemy session (as deault on SQLite), version aware 

expose = object() # a default decorator to add handlers


import sys
import re
import eyed3

from .. import path
from ..utils import obj2xml


def do():


    print("Modify a folder of mp3 files' ID3 based on file names")

    p = path(sys.argv[1] if len(sys.argv) > 1 else '.')

    print("Folder: {}".format(p))

    files = []

    for f, name in p.filter(filetype='mp3', recursive=False):

        x = re.split(' - ', name)

        album, artist, song = *((3-len(x)) * (None, )), *x

        print(album, artist, song)

        af = eyed3.load(f)
        print(af.tag)
        # af.tag.remove(f)

        # print(af.tag.artist, af.tag.album, af.tag.title)

    print("{} altogether".format(len(files)))


    # audiofile = eyed3.load("song.mp3")
    # audiofile.tag.artist = u"Integrity"
    # audiofile.tag.album = u"Humanity Is The Devil"
    # audiofile.tag.album_artist = u"Integrity"
    # audiofile.tag.title = u"Hollow"
    # audiofile.tag.track_num = 2

    # audiofile.tag.save()

    # audiofile.tag.remove()







if __name__ == '__main__': do()


"""


"""

from functools import wraps

from flask import current_app, url_for

from flask_login import (
    LoginManager as OrigLoginManager, current_user, UserMixin,
    login_required, login_user, logout_user
    )

from collections import OrderedDict as odict

from ..types import naodict

class LoginManager(OrigLoginManager):
	"""
	FLASK-LOGIN enhancement
	"""

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        self.custom_auth_callback = None
        self.nopriv_callback = None


    def custom_auth_handler(self, callback):
        self.custom_auth_callback = callback
        return callback

    def nopriviledge_handler(self, callback):
        self.nopriv_callback = callback
        return callback

    def nopriviledge(self):
        return self.nopriv_callback()

    @property
    def auth_map(self):

        am = []

        for endpoint, args, kwds in _auth_map:

            ok = self.custom_auth_callback(endpoint, args, kwds)

            # add only kwds to be used separately
            am.append((endpoint, ok, kwds))

        return am


    def get_menu_base(self):

        def has_no_empty_params(rule):
            defaults = rule.defaults if rule.defaults is not None else ()
            arguments = rule.arguments if rule.arguments is not None else ()
            return len(defaults) >= len(arguments)

        links = {}

        for rule in current_app.url_map.iter_rules():

            # Filter out rules we can't navigate to in a browser
            # and rules that require parameters
            if "GET" in rule.methods and has_no_empty_params(rule):
                url = url_for(rule.endpoint, **(rule.defaults or {}))
                key = url.split('/')[1:] # more questions
                # no duplicates allowed by werkzeug
                links[rule.endpoint] = (key, url)

        mp = []

        for endpoint, ok, kwds in self.auth_map:

            if ok and endpoint in links:
                key, url = links[endpoint]

                item = naodict()
                item.url = url
                item.key = key
                item.endpoint = endpoint
                item.subs = []

                mp.append(item)

        return mp

    @property
    def menu(self):

        def treebuilder(data, key=None, max_level=None):
            """
            TODO implement tree flattening after a certain level
            """

            # collectiong the items in order for this level
            col = odict()

            level = 0 if key is None else len(key)
            start = '' if key is None else '.'.join(key)

            # print('treebuilder', key, level, start)

            # check all the items
            for item in data:

                # if they are not for this node then nothing happens
                if not '.'.join(item.key).startswith(start):
                    # print("not considered - different branch", item)
                    continue

                if len(item.key) == level:
                    # print("not considered - current node")
                    continue

                # print("consider", item)
                # node to consider
                key_piece = item.key[level]

                # add to collection if not already there
                if key_piece not in col:
                    # create a pseudo-node
                    x = naodict()
                    x.url = "#"
                    x.subs = []
                    # x.endpoint = ''
                    x.key = item.key[:level+1]
                    col[key_piece] = x

                # build tree if not yet exist - triggered on any node
                if len(item.key) > level and not item.subs:
                    col[key_piece].subs = treebuilder(data, item.key[:level+1])

                # a real subnode -> update
                if len(item.key) == level+1:
                    # save subs
                    item.subs = col[key_piece].subs
                    # update with the real item
                    col[key_piece] = item
                    # print("added", item)

            # return only the subs in order
            return list(col.values())

        mp = self.get_menu_base()

        menu = naodict()
        menu.items = treebuilder(mp)

        return menu


# context problem might need to set app current_menu context...
_auth_map = []


# add to utils near login_required
def custom_auth(*args, **kwds):

    # if callable(args[0]): pass

    def decorator(func):

        # log.info("FUNC DECORATED: {}".format(func.__name__))
        # False means needs authentication

        _auth_map.append((func.__name__, args, kwds))

        @wraps(func)
        def decorated_view(*f_args, **f_kwds):

            lm = current_app.login_manager

            # add standard checks from flask-login.utils login required
            if lm.custom_auth_callback and lm.custom_auth_callback(func.__name__, args, kwds):
                return func(*f_args, **f_kwds)
            elif current_user.is_authenticated:
                # no right to access - redirect back to previous page 
                return lm.nopriviledge()
            else:
                return lm.unauthorized()

        return decorated_view

    return decorator

auth = custom_auth

# could be a lm method
def check_auth(*args, **kwds):

    lm = current_app.login_manager

    # call `custom_auth_callback` with empty function name
    return lm.custom_auth_callback \
        and lm.custom_auth_callback('', args, kwds)

check = check_auth


"""
Universal tools enhancing Python

Not related to any third party packages

The aim is to contribute these pieces into other
major projects
"""
import os
import re
import sys
import time
import pathlib
import contextlib
import collections
import yaml as pyyaml


__version__ = '0.2.11'


@contextlib.contextmanager
def chdir(path):
    """Change directory temporarily"""
    old_dir = os.getcwd()
    os.chdir(str(path))
    try:
        yield path
    finally:
        os.chdir(old_dir)


@contextlib.contextmanager
def benchmark(arg):
    """Measure time inside with statement"""
    # TODO use log instead
    print("Benchmark <{}> started...".format(arg))
    begin = time.time()
        
    try:
        yield 
    finally:
        last = human_time(time.time() - begin)
        print("Benchmark <{}> finished in: {}".format(arg, last))


def human_time(seconds):
    """Format seconds human readably

    TODO use `babel.dates.format_timedelta` with `datetime.timedelta(seconds=...)`
    """
    t = []
    
    if seconds > 60*60:
        hours, seconds = divmod(seconds, 60*60)
        t.append("{:.0f}h".format(hours))
    
    if seconds > 60:
        mins, seconds = divmod(seconds, 60)
        t.append("{:.0f}m".format(mins))

    t.append("{:.3f}s".format(seconds))

    return ' '.join(t)



# import time
# import functools
# import contextlib


# # TIMING CONTEXTMANAGER AND DECORATOR
# def benchmark(arg):

#     # print('ba', arg, callable(arg))

#     # normal use returns a context manager
#     if not callable(arg):
#         return Benchmark(str(arg))

#     # decorator use: wrapped into a benchmark context
#     func = arg
#     @functools.wraps(func)
#     def wrapper(*args, **kwds):
#         with Benchmark('FUNC ' + func.__name__):
#             return func(*args, **kwds)

#     return wrapper

# class Benchmark(object):
#     """
#     Based on something similar found long time ago on the internet
#     """
#     def __init__(self, name):
#         self._name = name
#         self._time = None

#     def __enter__(self):
#         self._begin = time.time()
#         print("Benchmark <{}> started ...".format(self._name))
          # return self

#     def __exit__(self, exc_type, exc_value, traceback):
#         self._time = time.time()-self._begin
#         print("Benchmark <{}> finished in: {}".format(self._name, self.time))
#         return False

#     @property
#     def time(self, raw=False):
#         if time is None:
#             raise ValueError('Not run yet')

#         return self._time if raw else human_time(self._time)
    





# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Recursive dictionary and list update
# TODO add support for list append
# move this to separate python package and import it from there


def recursive_update(base, other, recurse_lists=True):
    """
    Recursively updates base with other.

    Appends to lists as well.

    Returns the updated base data structure

    """
    if dictlike(base, other):
        for key, value in other.items():
            if key in base:
                base[key] = recursive_update(base[key], value)
            else:
                base[key] = value

    if listlike(base, other):
        if recurse_lists:
            for val in other:
                try:
                    idx = base.index(val)
                except ValueError:
                    base.append(val)
                else:
                    base[idx] = recursive_update(base[idx], val)
        else:
            base.extend(other)

    if setlike(base, other):
        for val in other:
            base.add(val)

    return base


def dictlike(*candidates):
    """checking"""
    for c in candidates:
        if not isinstance(c, (dict,)): return False
    return True

def listlike(*candidates):
    """checking"""
    for c in candidates:
        if not isinstance(c, (list,)): return False
    return True

def setlike(*candidates):
    """checking"""
    for c in candidates:
        if not isinstance(c, (set,)): return False
    return True



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# OrderedDict YAML version
# from https://github.com/wimglenn/oyaml
# hopefully it will soon gets integrated into pyYAML
# Python 3.7 supports ordered dict natively
# check for native C YAML lib for optimization
# v0.2.1

_items = 'viewitems' if sys.version_info < (3,) else 'items'

def map_representer(dumper, data):
    return dumper.represent_dict(getattr(data, _items)())


def map_constructor(loader, node):
    loader.flatten_mapping(node)
    return collections.OrderedDict(loader.construct_pairs(node))


if pyyaml.safe_dump is pyyaml.dump:
    # PyYAML v4.1
    SafeDumper = pyyaml.dumper.Dumper
    DangerDumper = pyyaml.dumper.DangerDumper
    SafeLoader = pyyaml.loader.Loader
    DangerLoader = pyyaml.loader.DangerLoader
else:
    SafeDumper = pyyaml.dumper.SafeDumper
    DangerDumper = pyyaml.dumper.Dumper
    SafeLoader = pyyaml.loader.SafeLoader
    DangerLoader = pyyaml.loader.Loader

pyyaml.add_representer(dict, map_representer, Dumper=SafeDumper)
pyyaml.add_representer(collections.OrderedDict, map_representer, Dumper=SafeDumper)
pyyaml.add_representer(dict, map_representer, Dumper=DangerDumper)
pyyaml.add_representer(collections.OrderedDict, map_representer, Dumper=DangerDumper)


if sys.version_info < (3, 7):
    pyyaml.add_constructor('tag:yaml.org,2002:map', map_constructor, Loader=SafeLoader)
    pyyaml.add_constructor('tag:yaml.org,2002:map', map_constructor, Loader=DangerLoader)


del map_constructor, map_representer


# public function overrides
def yaml_load(stream):
    if isinstance(stream, pathlib.Path):
        return pyyaml.load(stream.open(), Loader=SafeLoader)
    return pyyaml.load(stream, Loader=SafeLoader)

def yaml_dump(data, stream=None, **kwds):
    kwds.setdefault('default_flow_style', False)
    if isinstance(stream, pathlib.Path):
        with stream.open('w') as x:
            return pyyaml.dump_all([data], x, **kwds)
    return pyyaml.dump_all([data], stream, **kwds)

def yaml_print(*args, **kwds):
    if len(args) == 1:
        print(yaml_dump(args[0]))
    else:
        print(yaml_dump(args))
    if kwds:
        print(yaml_dump(kwds))



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Convert to under_score style


u1reg = re.compile(r'((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
u2reg1 = re.compile(r'(.)([A-Z][a-z]+)')
u2reg2 = re.compile(r'(a-z0-9])([A-Z])')

def underscore(text, method='single_reg'):
    """
    Largely based on Stack Overflow.
    Two methods are only implemented out of curiosity.

    Converts `CamelCase` or `camelCase` to under_score style.
    Able to handle `camelAndHTTPResponse` as `camel_and_http_response`.
    Avoids multiple underscores, so `under_Score` remains `under_score`
    does not become `under__score`.
    """

    if method == 'single_reg':
        return u1reg.sub(r'_\1', text).lower().replace('__', '_')

    if method == 'double_reg':
        temp = u2reg1.sub(r'\1_\2', text)
        return u2reg2.sub(r'\1_\2', temp).lower().replace('__', '_')

    raise ValueError("Method not recognized: {}".format(method))


# replace all fancy characters with underscore

def to_filename(name):
    return re.sub(r'[^\w\-_\.]', '_', str(name))


# padding strings
def padded(string, spaces=4):

    return '\n'.join([' '*spaces+l for l in string.split('\n')])

# flatten Python data structure
def flatten(d, parent_key='', sep='.'):
    """
    FROM http://stackoverflow.com/questions/6027558/flatten-nested-python-dictionaries-compressing-keys
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def unflatten(d):
    """
    TODO implement
    """



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Quaternion implementation

# import numpy as np

# class Quaternion:
#     """Quaternions for 3D rotations"""
#     def __init__(self, x):
#         self.x = np.asarray(x, dtype=float)
        
#     @classmethod
#     def from_v_theta(cls, v, theta):
#         """
#         Construct quaternion from unit vector v and rotation angle theta
#         """
#         theta = np.asarray(theta)
#         v = np.asarray(v)
        
#         s = np.sin(0.5 * theta)
#         c = np.cos(0.5 * theta)
#         vnrm = np.sqrt(np.sum(v * v))  # np.linalg.norm ???

#         # print([c], s*v/vnrm)

#         q = np.concatenate([[c], s * v / vnrm])
#         # print('raw data', q)
#         return cls(q)

#     def __repr__(self):
#         return "Quaternion:\n" + self.x.__repr__()

#     def __mul__(self, other):
#         # multiplication of two quaternions.
#         prod = self.x[:, None] * other.x

#         return self.__class__([(prod[0, 0] - prod[1, 1]
#                                  - prod[2, 2] - prod[3, 3]),
#                                 (prod[0, 1] + prod[1, 0]
#                                  + prod[2, 3] - prod[3, 2]),
#                                 (prod[0, 2] - prod[1, 3]
#                                  + prod[2, 0] + prod[3, 1]),
#                                 (prod[0, 3] + prod[1, 2]
#                                  - prod[2, 1] + prod[3, 0])])

#     def as_v_theta(self):
#         """Return the v, theta equivalent of the (normalized) quaternion"""
#         # compute theta
#         norm = np.sqrt((self.x ** 2).sum(0))
#         theta = 2 * np.arccos(self.x[0] / norm)

#         # compute the unit vector
#         v = np.array(self.x[1:], order='F', copy=True)
#         v /= np.sqrt(np.sum(v ** 2, 0))

#         return v, theta

#     def as_rotation_matrix(self):
#         """Return the rotation matrix of the (normalized) quaternion"""
#         v, theta = self.as_v_theta()
#         c = np.cos(theta)
#         s = np.sin(theta)

#         return np.array([[v[0] * v[0] * (1. - c) + c,
#                           v[0] * v[1] * (1. - c) - v[2] * s,
#                           v[0] * v[2] * (1. - c) + v[1] * s],
#                          [v[1] * v[0] * (1. - c) + v[2] * s,
#                           v[1] * v[1] * (1. - c) + c,
#                           v[1] * v[2] * (1. - c) - v[0] * s],
#                          [v[2] * v[0] * (1. - c) - v[1] * s,
#                           v[2] * v[1] * (1. - c) + v[0] * s,
#                           v[2] * v[2] * (1. - c) + c]]) 


# -*- coding: utf-8 -*-
"""
config

"""
from __future__ import division, print_function, unicode_literals

import os, sys, logging, logging.handlers, yaml, codecs, re, cherrypy

from docutils.core import publish_parts

from future.utils import iteritems

################################
# MISC HELPER FUNCTION

def path(*plus):
    # print(os.path.dirname(__file__))
    return os.path.abspath(os.path.join(os.path.dirname(__file__), *plus))

root_path = path('..') # the repository root from which diofa could be imported

# MISC HELPER FUNCTION
################################



################################
# CONFIG SETUP


class RecursiveDictionary(dict):
    """RecursiveDictionary provides the methods rec_update and iter_rec_update
    that can be used to update member dictionaries rather than overwriting
    them.

    Created 2009-05-20 by Jannis Andrija Schnitzer (jannis@itisme.org)
    https://gist.github.com/114831
    Modified 2012-12-01 by Szabolcs Blaga (szabolcs@syrus.hu)
    """
    def recurse(self, other, **third):
        """Recursively update the dictionary with the contents of other and
        third like dict.update() does - but don't overwrite sub-dictionaries.

        Example:
        >>> d = RecursiveDictionary({'foo': {'bar': 42}})
        >>> d.rec_update({'foo': {'baz': 36}})
        >>> d
        {'foo': {'baz': 36, 'bar': 42}}
        """
 
        self.iter_recurse(iteritems(other))
        self.iter_recurse(iteritems(third))

    def iter_recurse(self, iterator):
        for key, value in iterator:
            if key in self and isinstance(self[key], dict) and isinstance(value, dict):
                self[key] = RecursiveDictionary(self[key])
                self[key].recurse(value)
                self[key] = dict(self[key]) # avoid using multi-level recursive dictionaries
            else:
                self[key] = value

    def __repr__(self):
        return super(self.__class__, self).__repr__()


config = RecursiveDictionary()
for f in [path(n+'.yaml') for n in ('global', 'local')]:
    if os.path.isfile(f):
        c = yaml.safe_load(codecs.open(f, 'r', encoding='utf8'))
        if c is not None: config.recurse(c)

config = dict(config)

# CONFIG SETUP
################################


################################
# SET UP LOGGING

#log_path = path('..', 'temp', 'log')
#try: os.makedirs(log_path)
#except OSError: pass

DETAILS =   logging.Formatter('%(asctime)s - %(name)s %(levelname)s - %(message)s')
CONSOLE =   logging.Formatter('%(levelname)-8s - %(message)s')
# THREADS =   logging.Formatter('{levelname:1} {threadName:>10}: {message}', style='{') # Python 3
THREADS =   logging.Formatter('%(threadName)-10s %(name)-20s %(lineno)-4s %(asctime)s %(levelname)-7s %(message)s')
THREADS_A =   logging.Formatter('%(levelname)-7s %(threadName)-10s: %(message)s')
THREADS_B = logging.Formatter('<%(levelname)-7s %(threadName)-10s %(asctime)s - %(name)s>\n%(message)s')
FULL =      logging.Formatter('<%(levelname)-7s %(threadName)-10s %(asctime)s - %(name)s %(ip)-15s %(user)-4s>\n%(message)s')
#logging.basicConfig(format=console)
#logging.basicConfig(
#    level=logging.DEBUG,
#    format='%(levelname)-7s %(threadName)-10s %(asctime)s - %(name)s:\n\t%(message)s',
#)


class ConnInfo:
    """
    move to server, than no need to import cherrypy here
    """

    def __getitem__(self, name):
        """
        To allow this instance to look like a dict.
        """
        if name == "ip":
            try:
                result = cherrypy.request.remote.ip
            except:
                result = "<no:ip>"
        elif name == "user":
            try:
                result = cherrypy.user.id
            except:
                result = "<no:user>"
        else:
            result = self.__dict__.get(name, "?")
        return result

    def __iter__(self):
        """
        To allow iteration over keys, which will be merged into
        the LogRecord dict before formatting and output.
        """
        keys = ["ip", "user"]
        keys.extend(self.__dict__.keys())
        return keys.__iter__()


def logger(name):
    """returns the default logger for the modul"""
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    # log = logging.LoggerAdapter(log, ConnInfo())

    #~ TODO: log.addHandler(x)
    #~ log.addHandler(x)

    return log

# diofa default handler
# replace stdout to avoid conflict with citparser
_mdh = logging.StreamHandler(sys.stderr) # sys.stdout used to work perfectly with Apache...

debug = config['server'].get('debug', True)
if debug:
    print('DEBUG mode, log at DEBUG level, DB will be regenerated')
    _mdh.setLevel(logging.DEBUG)
else:
    print('Normal mode, log at INFO level')
    _mdh.setLevel(logging.INFO)

_mdh.setFormatter(THREADS)


root_log = logger(__name__) # root logger for diofa (every diofa logger inherits these settings)
    # logger('') - the real root logger - is not used because it would duplicate with cherrypy logging
    # the real root could be used if you remove the cherrypy handlers
    # BgTaskManager uses root_log to add/remove thread specific handlers
root_log.addHandler(_mdh)




log = root_log # specific logger for this root module (the same as root_log)


# error_file = os.path.join(log_path, 'error.log')
# info_file = os.path.join(log_path, 'info.log')
# debug_file = os.path.join(log_path, 'debug.log')

# log = logging.getLogger()
# log.setLevel(logging.DEBUG)



# error = logging.FileHandler(error_file)
# error.setLevel(logging.ERROR)
# error.setFormatter(details)

# # info = logging.handlers.TimedRotatingFileHandler(info_file, when='d', interval=1)
# # info.setLevel(logging.INFO)
# # info.setFormatter(details)

# debug = logging.handlers.TimedRotatingFileHandler(debug_file, when='d', interval=1)
# debug.setLevel(logging.DEBUG)
# debug.setFormatter(details)

# std = logging.StreamHandler(sys.stdout)
# std.setLevel(logging.DEBUG)
# std.setFormatter(console)

# log.addHandler(error)
# # log.addHandler(info)
# log.addHandler(debug)
# log.addHandler(std)

#logger = logging.getLogger

# log = logger(__name__)
# log.debug('hőúőúóüü')

# SETUP LOGGING
################################



################################
# POST PROCESSING CONFIG
# with docutils, log is already available here

def prepare_text(dic):
    "Itt inkább valamiféle rst-féle parsolás kellene"
    def icon_replace(match):
        # log.debug("Found icon: {0}".format(match.group(0)))
#        for g in match.groups():
#            log.debug('Icon element {0}'.format(g))

        x = '<i class="{0}"></i>'.format(" ".join(['icon-'+g for g in match.group(1).split(' ') if g]))
        # log.debug(x)
        return x


    for k,v in iteritems(dic):
        if isinstance(v, dict):
            dic[k] = prepare_text(v)
        else:
            t = publish_parts(v,writer_name='html')['html_body']
            # log.debug("Searching for icon tags...")
            t = re.sub(
                '\{([^{}]*)\}',
                icon_replace,
                t
            )

            dic[k] = t

    return dic

text = prepare_text(config.get('text', {}))


# print(text)

# log.debug("The recursively merged configuration:\n{0}\n".format(yaml.dump(config, default_flow_style=False)))

# POST PROCESSING CONFIG
################################














# -*- coding: utf-8 -*-
"""
this should be a basic package that
reads everything from a directory
structure (data lazy) and provide
a flexible object model to access
all the resources


- data files (rst, yaml, csv, xlsx, etc)
    rst needs support for :ref: inside nao datastructure
    add rst parsing to yaml files, also support cross yaml referencing with rst style tags
- database connections (sqlite, special yaml files using, sqlalchemy models?)
- automatically serve those files as a webserver
- python files for code support (no relative import only through nao.self?)
    - decorators to provide functionality outside (e.g. auth, etc)
- template files made available automatically (in nao.current.template)

later multilayer support should be added
the main resaon for that is supporting
multilingualism (also production, testing separation)

multilingualism is supported by babel negotiatble tags(e.g. en_US)

mode separation is set up based on _identifier style tags

"""
from __future__ import unicode_literals # no need in python3 stack

# basic logging
import os
import sys
import logging
import time
import functools
import codecs
import re
import collections

# from future.utils import python_2_unicode_compatible

from .utils import obj2xml, recursive_update, padded
from .types import odict, naodict, nao

from ._yaml import load as load, dump as dump

# get rid of __pycache__ for all
sys.dont_write_bytecode = True

op = os.path

import pathlib


def _load(source=None):
    """
    main entry point, loads everything from a given point
    of the file system (usually __file__)

    builds a tree based on file names and parse
    files based on extension and alphabetical
    order

    .yaml
        regular python data
        with some type of nao reference type available
        maybe all text fields interpreted as rst is a good idea
        add protocol handling as well (file://, ext://, oracle://, ...)
        see `logging.config.dictConfig` for reference

    .rst
        parsed as in sphinx docs into raw html
        :nao:`something` should be available

    .csv, .sql, .xls, .xlsx
        deferred datatable
        sql should run on the default db connection
        defined by nao-tree (with special .nao files)
        if nothing defined .nao-s own default sqlite
        db should be used (cashing, etc also in that)
    
    .nao
        special resource described in YAML
        e.g. database connaction, RPC service, aPI calls, etc.
        return a Python object???
        similar .rst's and .yaml's ext directive

    .html, .jinja
        parsed as jinja templates by default
        the object returned in load is available by the
        name nao in tempate global (think it over more)

    .py
        parsed as python code and an object tree
        is build (I know this is going to be controversial...)

    .jpg, .png, .mp3, .wav, .avi, ...
        parsed as media resource (embed in template)


    read system information as well and do
    everything in a multilayer and cascading
    manner

    returns a root element
    """
    if source is None:
        # doubles the check for existence, but avoids strange errors
        if op.exists('nao'):
            source = 'nao'

        # TODO check for nao.yaml, nao/, config.yaml, config/
        # in some order defined as a cutomizable parameter

        # finally load current directory as a config directory


    if isinstance(source, pathlib.Path):

        with source.open() as f:

            return yaml_load(f)
            

    if op.exists(source):
        if op.isdir(source):
            # log.debug("Loading from directory recursively")
            pass
            # TODO implement recursive directory loading
            # see Naopath.filter
        else:
            # log.debug("Loading from single file")
            with codecs.open(source, encoding='utf8') as f:
                # TODO add content negotiation here
                return yaml_load(f)

    else:
        # log.debug("Loading from string")

        # TODO create yaml naodictloader

        return yaml_load(source)

    # if source == None, then try to load [cwd]/nao

    # get directory based on source

    # create root_node of current directory
    # walk recursively, skip everything that starts with _
    # create nao* object-tree

    # nao.path(data.config.general)
    # str(data.config) # returns a path if it is a file/directory
    # nao.url(naobject :)





# no handler is added
rootlogger = logging.getLogger()

def logger(name,
        level='debug',
        add_default_handler=False,
    ):
    """
    For logging in modules outside of nao use::
    log = nao.logger(__name__)
    """
    # process dict config

    # not necessary since Python2.7 accept string levels
    levels = {
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }
    lvl = levels[level.lower()]

    log = logging.getLogger(name)
    log.setLevel(lvl)

    # optional only if there is no other logger
    if add_default_handler:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        fmtr = logging.Formatter('{levelname:1s} {message:s}', style='{')
        ch.setFormatter(fmtr)
        log.addHandler(ch)

    log.root = rootlogger

    return log

# default logger for nao
log = logger(__name__, 'debug', True)



from sqlalchemy.types import TypeDecorator, String, Integer, UnicodeText

from .. import yaml_load, yaml_dump


class MAC(TypeDecorator):
    """Represents a MAC address in 6 bytes.

    """
    impl = String(6)

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        return ''.join([chr(int(i, 16)) for i in value.split(':')])

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        
        return ':'.join(['{:0>2x}'.format(ord(i)).upper() for i in value])


class IP(TypeDecorator):
    """Represents in IPv4 address as an integer

    """
    import struct

    impl = Integer

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        if isinstance(value, int):
            return value
        if isinstance(value, str):
            return struct.unpack('I', bytes([v for v in value.split('.')]))[0]

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        
        return '.'.join([int(v) for i in struct.pack('I', value)])


class HexID(TypeDecorator):
    """Represents an 8 digit hexadecimal ID

    """
    impl = Integer

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        return int(value, 16)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        
        return '{:0>8x}'.format(value).upper()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# SQLAlchemy custom types


from sqlalchemy.types import TypeDecorator, UnicodeText
from .. import yaml_dump, yaml_load

class YAML(TypeDecorator):
    """Represents an 8 digit hexadecimal ID

    """
    impl = UnicodeText

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        return yaml_dump(value)


    def process_result_value(self, value, dialect):
        if value is None:
            return value
        
        return yaml_load(value)


from sqlalchemy.types import TypeDecorator, Text
import io

class DF(TypeDecorator):
    """Represents a Pandas DataFrame object as a csv text

    """
    import pandas as pd

    impl = Text  # UnicodeText might be better (pandas encodes?)

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if isinstance(value, pd.DataFrame):
            return value.to_csv(index=False)
        
        raise ValueError('need Pandas DataFrame or None')


    def process_result_value(self, value, dialect):
        if value is None:
            return value
        
        return pd.read_csv(io.StringIO(value))



# somecolumn.op('&')(1) == 1

class Mask(Integer):
    """implement a binary mask based on int
    
    use the like operator to decide weather other
    is contained in the mask
    """
    class comparator_factory(Integer.Comparator):
        def like(self, other, escape=None):
            print('USED', self, other)
            return self.op("&")(other) == other  # a & b == b  <-- b contained in a



# Werkzeug-based password hash (Werkzeug is included in Flask)
from werkzeug.security import generate_password_hash, check_password_hash
class Password(str):

    def __eq__(self, other):
        # print('CAAAAALLLLLEEEEEEEDDDDDD', self, other)

        # print('COMP', check_password_hash(generate_password_hash('valami'), 'valami'))
        return check_password_hash(self, other)


class PasswordHash(TypeDecorator):

    impl = Text

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if isinstance(value, str):
            return generate_password_hash(value)
        
        raise ValueError('need str or None')


    def process_result_value(self, value, dialect):
        
        print('proc res val', value, dialect)
        if value is None:
            return value
        
        return Password(value)



    # class comparator_factory(Text.Comparator):
    #     def __eg__(self, other):
    #         print('CAAAAALLLLLEEEEEEEDDDDDD')
    #         return check_password_hash(self, other)

    # class comparator_factory(Text.Comparator):
    #     def __eq__(self, other):
    #         return check_password_hash(self, other)


    # def compare_values(self, x, y):
    #     """determine that the password has changed"""
    #     return check_password_hash(x, y)


# @event.listens_for(User.password, 'set')
# def hash_pass(target, value, oldvalue, initiator):
#     target.password = 
#     available = [s['name'] for s in sensors]
#     enabled = [s['name'] for s in sensors if s['enable'] == True]
#     target.sensors_available = sum([2**i for i,n in enumerate(sensor_names) if n in available])
#     target.sensors_enabled = sum([2**i for i,n in enumerate(sensor_names) if n in enabled])
#     # return sensors

from sqlalchemy.types import TypeDecorator, Integer, Enum

class PythonEnum(TypeDecorator):
    """
    This might be accomplished via sqlalchemy standard enum


    """
    impl = Enum

    def __init__(self, enumtype, **kwds):
        super().__init__(*[e.name for e in enumtype], **kwds)
        self._enumtype = enumtype


    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        if isinstance(value, self._enumtype):
            return value.name
        if isinstance(value, int):
            return self._enumtype(value).name
        if isinstance(value, str):
            return self._enumtype[value].name

        # consider it an enum member
        return value.name


    def process_result_value(self, value, dialect):
        if value is None:
            return value
        # additional check constraint...
        return self._enumtype[value].name


    class comparator_factory(TypeDecorator.Comparator):
        """override to allow int comparisons"""

        def __eq__(self, other):

            print('EQ', self, type(self), other)

        def __le__(self, other):

            print('LE', self, type(self), other)

        def __ge__(self, other):

            print('GE', self, type(self), other)

        # def operate(self, op, other):

        #     print('COMPARE', self, type(self), op, other)
        #     print(dir(self))

        #     return True

    # #         if isinstance(value, int):
    # #             o = self._enumtype(value).name
    # #         if isinstance(value, str):
    # #             o = self._enumtype[value].name

    # #         return op(self, o)

    #     def reverse_operate(self, op, other):

    #         print('RCOMPARE', self, type(self), op, other)





# # store arbitrary numpy structure
# from sqlalchemy.types import TypeDecorator, LargeBinary
# import numpy as np


# class NumpyType(TypeDecorator):
#   impl = LargeBinary

#   def process_bind_param(self, value, dialect):
#     return zlib.compress(value.dumps(), 9)

#   def process_result_value(self, value, dialect):
#     return np.loads(zlib.decompress(value))


import zlib
import pickle
from sqlalchemy.types import TypeDecorator, LargeBinary

class PythonData(TypeDecorator):
    impl = LargeBinary

    def process_bind_param(self, value, dialect):
      return zlib.compress(pickle.dumps(value))
      # use 1 for best speed, 9 for best compression

    def process_result_value(self, value, dialect):
      return pickle.loads(zlib.decompress(value))




import warnings
import collections

class naodict(collections.OrderedDict):
    """basic attribute access for ordered dict

    keys that collide with predefined attribute names
    could not be set or accessed in attribute-style

    trying to set a predefined attribute produce `UserWarning`
    """
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        for key in self:
            if hasattr(self, key):
                warnings.warn(f"existing attribute will shadow attribute-style access for {key}")


    def __getattr__(self, key):
        # not called if the attribute present!
        if key in self:
            return self[key]
        return super().__getattr__(key)


    def __setattr__(self, key, value):
        if hasattr(self, key):
            warnings.warn(f"updating existing attribute {key}")
            super().__setattr__(key, value)
        else:
            self[key] = value


    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if hasattr(self, key):
            warnings.warn(f"existing attribute will shadow attribute-style access for {key}")



"""
Loader

ParamLoader
- parse params

RangeLoader
- parse ranges
- instantiate individual classes
- ?? resources

ResourceLoader
- maintain central read-only resource cache
  for all classes instantiated


instance-level:
TODO call this a Resource?


SQLResource
- load from SQL

ExcelResource
- load from Excel


class-level:

MixinLoader
- load for all loadable members
  of a Base class
- multi level, all model class could have
  their own loader definition


ProcedureLoader
- load by many different procedures


TreeLoader
- load from an inheritance tree
- multi level other classes in the same
  tree are also instantiated


InverseTreeLoader
- load from an inverse inheritance tree
- also multi-class

instance or class?
support multitreading by param instantiation
from ParamLoader onward

Loader is a virtual base-class


TODO

generalize behaviour
currently check, load, diff, implicit foreign
make these things more clear-cut!!!!
!!!!!!!!!!!!!!

TODO

Check should be loader-internal
diff should also be loader internal (implicit foreign aware)

implicit foreign is inter-loader and inter-modelclass

If check is inter-loader that is a separate process
Loading resources should also be centralized for caching


"""
import numpy as np
import decimal as dec
import itertools
from inspect import isclass, getmro


import sqlalchemy.types as types
from sqlalchemy.types import DateTime, Date
from sqlalchemy.types import Integer, BigInteger, Float, Numeric, Boolean, Enum
from sqlalchemy.types import String, Text, Unicode, UnicodeText
from sqlalchemy.schema import Column
from sqlalchemy.orm import sessionmaker
from sqlalchemy.inspection import inspect
from sqlalchemy.ext.declarative import declared_attr

from sqlalchemy_continuum import version_class

import pandas as pd


from . import logger, benchmark, odict

log = logger(__name__)


###########################
# allow sqlite to accept Decimal

import sqlite3

def adapt_decimal(d):
    return str(d)

def convert_decimal(s):
    return dec.Decimal(s)

# Register the adapter
sqlite3.register_adapter(dec.Decimal, adapt_decimal)

# Register the converter
sqlite3.register_converter("decimal", convert_decimal)




############################
# SQLAlchemy custom types


class Period(types.TypeDecorator):
    '''
    Adds SQLAlchemy support for pandas.Period type
    '''

    impl = types.String

    def process_bind_param(self, value, dialect):
        if type(value) == pd.Period:
            return str(value)
        else:
            return value

    def process_result_value(self, value, dialect):
        if value.startswith("Period"):  # TODO add more security
            return eval(value, {'Period': pd.Period})
        else:
            return pd.Period(value)

    def copy(self, **kw):
        return Period(self.impl.length)



# TODO need to centralize resource loading (query and Excel)
# to ensure data integrity


class LoaderError(Exception): pass


class LoadableBase():
    """
    TODO move this to MixinLoader
    that inherits from ParamLoader

    """

    @classmethod 
    def sorted_subclasses(cls):
        """
        the collecting mechanism is that the classes
        should be inherited from the caller

        TODO change that they should inherit from
        Base and have a __loader__ attribute
        create Base.metadata.load_all() somehow
        """

        # get all subclasses that should have a __tablename__:
        lst = odict()

        for c in cls.subclasses():
            # print('scls', c)
            if hasattr(c, '__tablename__'):
                lst[c.__tablename__] = last_cls = c
                

        # get metadata
        metadata = last_cls.get_metadata()
        
        loader_classes = odict()

        if metadata:
            for t in metadata.sorted_tables:
                if t.name in lst:
                    # print('lcls', c)
                    yield lst[t.name]

            return loader_classes

        else:
            raise ValueError("No metadata found :(")


    @classmethod 
    def subclasses(cls):
        yield cls
        for subclass in cls.__subclasses__():
            # yield subclass  # this is not necessary because of the above yield
            yield from subclass.subclasses()


    @classmethod 
    def get_metadata(cls):
        """
        _declared_class_registry might be safer...
        """
        for c in cls.__mro__:
            print('mro', c)
            if hasattr(c, 'metadata'):
                return c.metadata


    @classmethod 
    def sorted_tables(cls, versioned=False):

        # TODO SQLAlchemy-Continuum versioning is not included here!!!
        # use metadata somehow
        # or include version_class operator...
        for c in cls.sorted_subclasses():

            yield c.__table__

            # sqlalchemy_continuum extension
            if hasattr(c, '__versioned__'):
                yield(version_class(c).__table__)

    
    @classmethod
    def create_all(cls, engine):
 
        for t in cls.sorted_tables(True):
            t.create(engine, checkfirst=True)


    @classmethod
    def drop_all(cls, engine):
 
        for t in reversed(list(cls.sorted_tables(True))):
            t.drop(engine, checkfirst=True)


    @classmethod
    def recreate_all(cls, engine):
 
        cls.drop_all(engine)
        cls.create_all(engine)


    def data_update(self, other):
        """
        update a class with a class of the same type
        """
        for c in type(self).__table__.c:
            a = getattr(self, c.key)
            b = getattr(other, c.key)
            if a is None and b is not None:
                setattr(self, c.key, b)
                log.debug("Column {} updated: {} -> {}".format(c.key, a, b))

                   
    def data2dict(self):
        """
        generate dict from class
        """
        ret = odict()
        for c in type(self).__table__.c:
            ret[c.key] = a = getattr(self, c.key)
            log.debug("Column {}: {}".format(c.key, a))

        return ret




class LoadableMixin(LoadableBase):
    """
    This is the Basic mixin that performs loading functionality

    This should be created similarly to declarative base otherwise
    shared among all model classes in the same interpreter process

    """
    __loader__ = None  #: override class attribute in subclasses
    __loader_args__ = {} #: loading arguments shared by subclasses


    @classmethod
    def load(cls, params=None, ranges=None):
        """
        Loads data for one class
        """
        # loader_classes, metadata = [cls], get_metadata(cls)

        # for pars in cls._parse_ranges(ranges, params):

        #     s = cls._get_session(pars, metadata)

        #     res = cls._load(pars)

        #     s.bulk_insert_mappings(c, res)

        #     s.commit()
        #     s.close()

        cls.load_all(params, ranges)


    @classmethod
    @benchmark
    def load_all(cls, params=None, ranges=None):
        """

        `params`
        These params are used to prepare sql-statements 

        As sorted_subclasses yields the class itself single is
        not necessary anymore


        `ranges`
        Used to iterate through ranges (preferably an `odict`)
        """
        # TODO somehow stop relying on table names

        metadata = cls.get_metadata()

        # this should work for single class as well :)
        # TODO name it something like subclass load
        loader_classes = odict([(c.__tablename__, c) for c in cls.sorted_subclasses()])

        log.debug("loader classes: {}".format(loader_classes))
        log.debug("metadata: {}".format(metadata))

        for pars in cls._parse_ranges(ranges, params):

            # provide a session
            s = cls._get_session(pars, metadata)

            # need to perform checks on all classes
            # checks = cls._check_all(loader_classes, pars)
            # TODO should it be a TRUE or FALSE query?
            # TODO CAPITAL LETTER PARAMETERS HOW?

            # loads all data into cache
            cache = cls._load_all(loader_classes, pars)

            # need to diff here
            # VERY resource intensive (-> bulk insert almost unnecessary)
            diff = pars.pop("diff_load", False)

            # bulk insert chould operate separately by the cache
            for key, res in cache.items():
                c = loader_classes[key]
                s.bulk_insert_mappings(c, res)

            s.commit()
        
            # TODO implement row-by-row insert for debugging
            # for row in res:
            #     s.bulk_insert_mappings(self.cls, [row])
            #     s.commit()
            s.close()


    @classmethod
    def _load_inner(cls, params=None):
        """
        Get classes, no ranges, no session
        
        return cls objects loaded from engine

        """
        # TODO use only __loader__args__ instead
        # and initialize loader with the given params
        # pop.loader_name
        # update recursive (or flatten-unflatten) with current params

        if isinstance(cls.__loader__, Loader):
            # delegate task to loader object
            log.debug('Loading for model class: {}'.format(cls))

            return cls.__loader__.load(cls, params)

        else:
            # rasie some more sensible error
            log.info("{} is not loadable!".format(cls))
        


    @classmethod
    def _load_all(cls, loader_classes, params):
        """
        Get classes, no ranges, no session

        Implement implicit foreign keys
        """
        # TODO use cache only if there is implicit foreign
        cache = odict()
        implicit_foreign = []

        def get_pairs(relation):

            return tuple(zip(*[(a.name, b.name) for a,b in relation.local_remote_pairs]))

        def get_pks(obj):

            return tuple([key.name for key in inspect(obj).primary_key])            


        for tname, klass in loader_classes.items():

            # relies on _load_inner here
            log.debug("Loading classes for {} and {}".format(tname, klass))
            res = klass._load_inner(params)

            cache[tname] = res

            for r in inspect(klass).relationships:

                # TODO what if related class is not in loader classes
                # ** should be
                rtbl = r.target.name

                # check if target allows imlicit foreign key addition
                if (rtbl in loader_classes
                    and issubclass(loader_classes[rtbl], ImplicitForeignMixin)
                ):
                    # could be filtered here by the target needs to be a primary constraint thing
                    # target should already exist in the cache because of sorted_tables
                    implicit_foreign.append((klass, r))


                    

            # TODO For DiffLoader check existence here
            # New (add new items with insert_date)
            # Match (implicit records should drop, others update_date)
            # Delete (no actual deletion only delete_date)


        # making implicit foregin key happen
        # TODO many to many??? association proxies???
        # association proxies does not matter (just proxies)
        # many-to-many is difficult, but it is two 1-N in reality
        # ??? load N-N mid-table implicitly
        # TODO not null howto?

        log.debug("Searching implicit foreign keys...")

        def get_index(table, keys):
            """
            TODO might implement caching of indices here
            """
            index = []
            for row in cache[table.name]:
                index.append(tuple([row[k] for k in keys]))
            return index

        # need the reversed order for implicit foreign search
        # OR leave some classes out
        for klass, relation in reversed(implicit_foreign):
            

            # this is not necessarily foreign and primary it could be a backref
            aks, bks = get_pairs(relation)

            pks = get_pks(relation.target)

            if bks != pks:
                log.debug("Skipping relation from {} to {}".format(klass.__table__, relation.target))
                continue

            log.debug("Checking implicit foreign for {} on {}".format(klass, relation))
            
            log.debug("foreign key: {} | primary key: {}".format(aks, bks))
            # target is accepting new records with implicit_foreign flag
            target_index = get_index(relation.target, bks)
            # print(target_index)
            target_class = loader_classes[relation.target.name]
            log.debug("target class: {}".format(target_class))

            for row in cache[klass.__tablename__]:

                # prevent implicitly generated rows to be considered!
                # if IMPLICIT_FOREIGN in row: continue
                
                ind = tuple([row[k] for k in aks])

                if any([i is None for i in ind]): continue

                # print(ind)
                if ind not in target_index:

                    log.warning("Implicit Foreign {} ({}) from table {}".format(
                        relation.target.name,
                        ", ".join([str(i) for i in ind]),
                        klass.__tablename__
                    ))

                    new_row = dict((k,v) for k,v in zip(bks, ind))
                    new_row[IMPLICIT_FOREIGN] = True
                    cache[relation.target.name].append(new_row)

                    # update target_index
                    target_index.append(ind)
                    # target_class.add_implicit_foreign(ind) # return a dict...

            # read all data from klass and check if already in target_data

        return cache


    @classmethod
    def _parse_ranges(cls, ranges, params):

        if ranges is not None:

            names, iterables = ranges.keys(), ranges.values()

            # initialize all parameters once with a Carthesian product
            prod = itertools.product(*iterables)

            for vector in prod:
                
                d = dict(zip(names, vector))

                log.info("LOADING values: {}".format(d))

                p = params.copy()
                p.update(d)


                log.debug("loading params: {}".format(p))
                yield p


        # simply loading with params
        else:
            log.debug("loading params: {}".format(p))
            yield params



    @classmethod
    def _parse_data(cls, data,
        return_raw_dict=True,
        position_based=True,
        only_one_row=False,
        ):
        """
        returns a list of dicts

        validated and coerced to the corresponding model
        """

        # TODO remove columns with special usage
        columns = cls.__table__.columns
        log.debug("Parsing column {}".format(columns))

        res = []

        if only_one_row:
            data = [data]

        # TODO if slow then pandas might use some C magic for this but I doubt (should check)
        i, j = 0, 0  # if the resultproxy has no rows...

        for i,row in enumerate(data):
            # allow extra columns at the end with default values or null
            if position_based and len(row) > len(columns):
                raise ValueError("Degenarate raw (col count mismatch)")

            new_row = {}

            # log.debug("The {} raw row: {}".format(i, row))

            for j,col in enumerate(columns):

                if position_based:
                    # it is OK to have extra columns at the end
                    if j >= len(row): break
                    val = row[j]

                elif col.name in row:
                    # name based could have extra rows anywhere
                    val = row[col.name]
                else:
                    log.warning("Column {} missing for row {}".format(col.name, i))
                    continue

                if val is not None:
                    try:
                        val = cls._parse_value(val, col)
                    except:
                        log.error("row {}, col {}, value [{}], type {}".format(i, j, row[j], type(row[j])))
                        log.error("COLUMN {} ({}) pk: {}".format(col.name, col.type, col.primary_key))
                        raise

                # log.debug("Value parsed for {}: {}".format(col.name, val))
                new_row[col.name] = val

            if return_raw_dict:
                res.append(new_row)
            else:
                res.append(cls(**new_row))

        log.debug("Parsed {} rows with {} columns".format(i+1, j+1))

        if only_one_row:
            return res[0]
        else:
            return res


    @staticmethod
    def _parse_value(val, col):

        # TODO check foreign keys for uniqueness and no None!
        # if c.foreign_keys:
        #     log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))

        # TODO think about primary key column check
        # TODO think about unique constraint check

        # determine numerical input types
        raw_numeric_types = (int, dec.Decimal, float)

        # SQLalchemy types categorized
        numeric_types = (Integer, BigInteger, Numeric, Float)
        date_types = (Date, DateTime)
        string_types = (String, Unicode, Text, UnicodeText)
        bool_types = (Boolean, )

        # very lazy approach to bool casting
        true_vals = (True, 'Y', 'I', 1, '1')  
        # false_vals = (False, 'N', 'N', 0, '0')

        # numeric types only go to numeric types
        if isinstance(val, raw_numeric_types):
            if isinstance(col.type, numeric_types):
                pass
                # TODO check for possible precision loss
                # log.debug("Casting numeric to numeric")
                # val = dec.Decimal(str(val)) OR let the engine do that :)

            elif isinstance(col.type, string_types):
                log.warning("Casting numeric to string")
                orig = val
                val = str(val)
                if col.type.length is not None and len(val) > col.type.length:
                    val = val[:col.type.length]
                    # remove unnecessary end zeros...
                    if orig != type(orig)(val):
                        raise ValueError("COLUMN {} would have been truncated {}".format(col.name, val))


            else:
                raise ValueError("Casting numeric to non-numeric!")

        # str and anything except for numeric and None
        else: 
            if isinstance(col.type, date_types):
                val = pd.to_datetime(val)
                # TODO stp using pd
                # TODO unrealistic date (what is realistic? (today +/- 120 years)
            # series.between(left, right)

            elif isinstance(col.type, numeric_types):
                # Everything casted to Decimal to avoid precision problems, only one conversion occurs at the SQL side
                if val.startswith('0'):
                    log.warning("Converting '0' starting strig to numeric")
                # check dtypes and issue warnings
                val = dec.Decimal(val)

            elif isinstance(col.type, string_types):
                val = val.strip()
                # callback mechanism for maximum string length in column! (self._callback_string_length)
                # check string and unicode columns length attribute
                if col.type.length is not None and len(val) > col.type.length:
                    raise ValueError("COLUMN {} would have been truncated {}".format(col.name, val))


            elif isinstance(col.type, bool_types):
                val = val in true_vals

        return val


    @staticmethod
    def _get_session(params, metadata):

        te = params.pop('target_engine', None) or metadata.bind

        if not te:
            raise ValueError("no target engine")

        log.debug('Target engine: {}'.format(te))
        
        return sessionmaker(bind=te)()


IMPLICIT_FOREIGN = 'implicit_foreign'

class ImplicitForeignMixin(LoadableMixin):
    """
    This class is mainly an indicator to initiate
    impliciet loading mechanism based on the foreign keys
    of other `Loadable` classes of the same `LoadableBase`.

    Adds a `Column` by the name defined in `IMPLICIT_FOREIGN`
    constant.
    """





def anon(cls):
    return None


# TODO remove column from load iteration!
# PATCH using `declared_attr` moves it to the end otherwise
# `column_instance._create_order` defines or need to flag mixin
# columns otherwise


setattr(ImplicitForeignMixin, IMPLICIT_FOREIGN, declared_attr(lambda cls: Column(Boolean, default=False)))







def anon(cls):
    return None

# TODO these methods might be moved to a class definition
# from which LoadableMixin inherits

import pandas as pd  # TODO REMOVE pandas dependency!


class Loader(object):
    """
    Base class for all the loading functionality

    This could be an abstract base clas
    """
    pass


class ParamLoader(Loader):
    
    def __init__(self, **kwds):
        """
        """
        self.kwds = kwds


    def load(self, model_class, params):

        self._get_config(model_class, params)

        self.model_class = model_class
        self.table = model_class.__table__

        # debug usually from load or load_all call
        self.debug = self.cfg.pop('debug', False)

        self.tbl = self.table
        self.cls = self.model_class
        # log.debug(tbl)

        # call class specific loader mechanism
        return self._load()


    def _get_config(self, cls, params):
        x = cls.__loader_args__.copy()
        x.update(self.kwds)
        x.update(params)
        self.cfg = x




class SQLLoader(ParamLoader):


    @staticmethod
    def prepare_sql(sql, **params):
        sql = sql.strip()
        if sql.endswith(';'): sql = sql[:-1]
        sql = sql.format(**params)
        return sql


    def _load(self):

        cfg = self.cfg
        # get engine
        engine = cfg.get('source_engine')

        # INFO no defaults needed as if there is it should be in the
        # SQL (otherwise a cfg called params could be used)
        sql = cfg.get('load_sql')
        # create prepare sql class which use parameter inclusion
        # might wanna check sqlalchemy syntax for that!
        sql = self.prepare_sql(sql, **cfg)
        # log.debug(sql)
        
        # read sql
        res = self.model_class._parse_data(engine.execute(sql))
        # res = self._read_sql_slow(sql, engine)

        # search duplicates
        if self.debug:

            import pandas as pd

            df = pd.DataFrame(res)
            pks = [c.name for c in self.table.columns if c.primary_key]
            log.debug("primary key: {}".format(pks))
            dpl = df.duplicated(subset=pks, keep=False)
            if dpl.any():
                log.debug("DUPLICATE: {}".format(df.duplicated(subset=pks).any()))
                df[dpl].to_csv('dupidup.csv')
                # TODO make output filename more sane

        # TODO add callback mechanism for other tables
        # for c in self.cls.__table__.columns:
        #     if c.foreign_keys:
        #         log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))

        return res


    


class ExcelLoader(ParamLoader):
    

    def _load(self):
        """
        Excel load mechanism using pandas for quick implementation
        """

        return

        cfg = self.cfg

        path = cfg.get('path')
        sheet = cfg.get('sheet', None)
        cols = cfg.get('columns', None)


        import pandas as pd

        xl = pd.ExcelFile(path)
        # print(xl.sheet_names)

        df = xl.parse(sheet if sheet is not None else 0)

        # filter by cols here

        users = df.to_dict(orient='records')

        groups = {}

        ret = []

        def get_group(name):
            if name not in groups:
                groups[name] = Group(name=name)
            
            return groups[name]

        for u in users:

            c = dict(
                login_ldap = u['id'],
                email = u['email'],
                name_first = u['first'],
                name_last = u['last'],
                name_title = u['title'],
                name_nick = u['nick'],
                name_hanzi = u['hanzi'],
                tele_ext = str(u['ext']),
                tele_mobile = u['mobile'],
                # name_nick = u['nick'] if u['nick'] else None,
                )

            # c = cls(**c)

            # Membership(c, get_group(u['department']), u['order'])

            ret.append(c)

        return ret

class DiffLoader(ParamLoader):
    pass


class SQLDiffLoader(SQLLoader, DiffLoader):
    """
    Search for new items based on primary key

    Auto-add functionality if pk is missing
    (tricky it should be used from other table loaders!!!)
    TODO

    TODO Implement callback functionality to add missing entities
    """



class ComplexLoader(ParamLoader):

    """
    Loader is a class that all other loader classes
    should inherit from and could be given to a model
    class with a LoadableMixin mixin
    or only with `__loader__` attribute


    Loader child classes should be initialized with
    a varibale set of parameters and all classes
    that are inherited should be initialized with the
    same parameter set and loaded in mro order


    Every class has its own class-level registry of
    procedures that could be added with a decorator function
    and return dictionary like data.


    Resources and return values are handled centrally
    
    This structure is similar to the already implemented
    model-based loading mechanism but that constitutes
    a higher level implementation of the same logic
    UNIFY SOMEHOW???



    In this case there will be a complete loader structure
    attched to a data model table called (RiskData)


    """
    methods = odict()

    def _load(self, params, ranges):
        """
        Loads all subclasses in an MRO (?) order and
        and returns all the loading results in one
        batch (e.g.) to imput into a model-based loader

        Resources are cached based on resource location

        """

        resource_cache = {}
        results = odict()

        for n, m in methods.items():

            results[n] = m(**self.cfg)

        return results

    @classmethod
    def add(cls, name):

        def decorator(func):
            cls.methods[name] = func
            return func

        return decorator







#############################################################
# OLD LOADER based on Pandas
# It was not working mainly due to mixed types row handling difficulties

# @benchmark
# def _parse_sql(self, sql, engine):
#     """
#     TODO not in use anymore
    
#     Cannot be significantly faster as it is necessary to read data into memory
#     for conversions
#     """
#     # coerce_float=False ensures all dtypes are object_
#     with benchmark("Reading data into memory..."):
#         df = pd.read_sql(sql, engine, coerce_float=False)

#     # TODO create own describe based on pandas
#     # show more sophisticated data
#     log.debug("Describing dataframe before conversion:")
#     log.debug(df.describe(include='all'))

#     log.debug("Datatypes of dataframe:")
#     log.debug(df.dtypes)

#     # print(df)

#     col_names = [c.name for c in self.cls.__table__.columns]
#     df.columns = col_names
#     # TODO check compatibility of dimensions

#     index = []
#     for i, c in enumerate(self.cls.__table__.columns):

#         log.debug("COLUMN {} ({}) pk: {}".format(c.name, c.type, c.primary_key))
#         # df[c.name] = df[c.name].astype(str)

#         col = df[c.name] # working with one col at a time
#         col = col[pd.notnull(col)] # working with not None only

#         print(col)

#         # TODO check foreign keys for uniqueness and no None!
#         if c.foreign_keys:
#             log.debug("Foreign keys for {} are {}".format(c.name, c.foreign_keys))


#         if isinstance(c.type, (Date, DateTime)):
#             col = pd.to_datetime(col)
#             # TODO unrealistic date (what is realistic? (today +/- 120)
#             # series.between(left, right)

#         elif isinstance(c.type, (Integer, BigInteger, Numeric, Float)):
#             # Everything casted to Decimal to avoid precision problems, only one conversion occurs at the SQL side
#             if col.str.startswith('0').any():
#                 log.warning("Converting '0' starting strig to numeric")
#             # check dtypes and issue warnings
#             col = col.astype(dec.Decimal)
#             # pd.to_numeric(df[c.name], errors='coerce')  
#             # df.ix(pd.notnull(df[c.name]), c.name).astype(dec.Decimal) #
#             # df[c.name].replace(pd.nan, None, inplace=True)

#         elif isinstance(c.type, (String, Text, Unicode, UnicodeText)):
#             col = col.str.strip()
#             mx = col.str.len().max()  # .map(len).max() is quicker a bit
#             log.debug("COLUMN {} max length: {}".format(c.name, mx))
#             # check string and unicode columns length attribute
#             if mx > c.type.length:
#                 raise ValueError("COLUMN {} would have been truncated".format(c.name))


#         elif isinstance(c.type, (Boolean, )):
#             col = col.astype('bool')  # remove none!
#             # Allow Y-N, 0-1, True-False conversion explicitly

#         #   print(i.enums)

#         df[c.name] = col # write back parsed col
        

#     # to_numeric introduce np.nan -> monkey-patch:
#     # df = df.where(pd.notnull(df), None) # notnull selects the elements NOT replaced
#     # TODO move this to to_dict somehow as in a simple dict no np.nan should occur...

#     # show more sophisticated data
#     log.debug("Describing dataframe after conversion:")
#     log.debug(df.describe(include='all'))

#     log.debug("Datatypes of dataframe:")
#     log.debug(df.dtypes)

#     # print(df)
        
#     # convert to dictionary works only without using index
#     # df.set_index(index, inplace=True)

#     # custom df.to_dict
#     res = []
#     for df_row in df.itertuples():
#         r = {}
#         for i,c in enumerate(col_names):
#             r[c] = df_row[i]
#         res.append(r)

#     return res

#     # return df.to_dict(orient='records')  # avoid indices for this operation!





import codecs
import os
import re

import setuptools

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# long_description = read('README.rst')

# setup(
#   name = 'nao',
#   packages = ['nao'],  # find_packages(),
#   version = '0.2.1',
#   description = 'Intelligent data manipulation tools',
#   author = '',
#   author_email = 'szabolcs.blaga@gmail.com',
#   url = 'https://github.com/blagasz/nao',
#   download_url = 'https://github.com/blagasz/nao/tarball/0.1',
#   license = 'GPL',
#   install_requires=[
#     'PyYAML',
#     # 'numpy==1.11.1',
#     # 'pandas==0.18.1',  # for datetime conversion and distinct
#     # 'SQLAlchemy',
#   ],
#   keywords = ['data', 'yaml', 'multilingual', 'multivalue', 'config', 'flask', 'sqlalchemy'],
#   classifiers = [],
# )

# based on https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56


from distutils.core import setup

setup(
  name = 'nao',
  packages = ['nao', 'nao.ext'],
  version = find_version("nao", "__init__.py"),
  license='GPL',
  description = 'Intelligent data manipulation tools',
  author = 'Szabolcs Blága',
  author_email = 'szabolcs.blaga@gmail.com',
  url = 'https://github.com/blagasz/nao',
  download_url = 'https://github.com/blagasz/nao/archive/v0.2.zip',
  keywords = ['data', 'yaml', 'multilingual', 'multivalue', 'config', 'flask', 'sqlalchemy'],
  setup_requires=['wheel'],
  install_requires=[
    'PyYAML',
    # 'numpy==1.11.1',
    # 'pandas==0.18.1',  # for datetime conversion and distinct
    # 'SQLAlchemy',
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Utilities',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)